﻿/*
AwesomeNET by xeetsh - Maximilian T. Schmitt

Licensed under the Unity Asset Store Terms of Service and EULA. See https://unity3d.com/legal/as_terms for a complete version of the license

AwesomeNET Doccumentation: https://docs.awesomenet.xeet.sh
Contact/Support/Feature requests: https://xeet.sh/contact.php
Twitter: @xeetsh
*/

using System.Collections;
using System.Collections.Generic;
using UnityEngine;
using System.Net;
using System.Net.Sockets;
using System;
using System.IO;
using System.Text;
using System.Runtime.Serialization;
using System.Xml;
using System.Runtime.CompilerServices;
using System.Reflection;


public class NetworkController : MonoBehaviour
{
    public MatchmakingClient m_mmc;
    public GameClient m_gc;
    private List<string[]> m_commandQueue = new List<string[]>();
    private bool m_offlineGame = false;
    [HideInInspector] public bool m_connected = false;

    void Awake() {
        // Kill Networkmanager if there is another instance of a GameManager present
        if (FindObjectsOfType(typeof(NetworkController)).Length > 1) {
            Debug.Log("NetworkController: Already found instance of script in scene; destroying.");
            DestroyImmediate(gameObject);
        } else {
            DontDestroyOnLoad(gameObject);
        }
    }

    public void Start() {
        StartCoroutine(GSCommandLoop());
    }

    public void SendGSCommand(string objectID, string actionID, object[] argument) {
        if (!m_offlineGame)
            m_gc.Send(objectID, actionID, SerializeArgument(argument));
        else {
            // If the player is offline construct the argumentvector and queue it without sending it to any server
            string[] argv = new string[3];
            argv[0] = objectID;
            argv[1] = actionID;
            argv[2] = System.Text.Encoding.ASCII.GetString(SerializeArgument(argument));
            ExecuteGSCommand(argv);
        }
    }

    public void ExecuteGSCommand(string[] argv) {
        // Queue the command in for execution
        if (NetworkGameObjectConnector.m_connectors.ContainsKey(argv[0])) {
            m_commandQueue.Add(argv);
        } else {
            Debug.Log(String.Format("Couldn't run command because there is no object with the id {0}", argv[0]));
        }
    }

    private IEnumerator GSCommandLoop() {
        while(true) {
            if (m_commandQueue.Count != 0) {
                // Iterate through every queued command and run it
                List<string[]> deleteMe = new List<string[]>();
                foreach (string[] command in m_commandQueue) {
                    NetworkGameObjectConnector con = NetworkGameObjectConnector.m_connectors[command[0]];
                    foreach(NetworkCommand nc in con.m_commands) {
                        if (nc.actionID == command[1]) {
                            // Found the command to run! FIRE!

                            // Build a type array from a string array
                            Type[] types = new Type[nc.argumentType.Length];
                            for (int i = 0; i < types.Length; i++) {
                                types[i] = Type.GetType(nc.argumentType[i]);
                            }
                            
                            // build a method from a string
                            MethodInfo method = nc.component.GetType().GetMethod(nc.command);
                            object[] argument = DeserializeArgument(command[2], types);

                            // Finally FIRE!
                            method.Invoke(nc.component, argument);

                            Debug.Log(String.Format("Executed command: {0} for object: {1}", nc.action, con.m_objectID));

                            // Queue the command to be deleted
                            deleteMe.Add(command);
                            break;
                        }
                    }
                }
                // Remove ran commands
                foreach (string[] command in deleteMe) {
                    m_commandQueue.Remove(command);
                }
            }
            yield return null;
        }
    }

    public bool ConnectMatchmaking(string host, int port) {
        m_mmc = new MatchmakingClient(host, port);
        if (!m_connected && m_mmc.Connect()) {
            m_connected = true;
            return true;
        } else {
            return false;
        }
    }

    public bool DisconnectMatchmaking() {
        if (m_connected && m_mmc.Disconnect()) {
            m_connected = false;
            return true;
        } else {
            return false;
        }
    }

    public bool JoinGame(string id, string player) {
        if (m_connected) {
            IPEndPoint ep = m_mmc.JoinGame(id, player);
            // Spawn new game client
            m_gc = new GameClient(ep.Address.ToString(), ep.Port, this, id);
            if (m_gc.Connect()) {
                NetworkGameObjectConnector.m_connectors.Clear();
                NetworkGameObjectConnector.m_humanRedableIDs.Clear();
                return true;
            }
            Debug.Log("Couldn't start a new game. Maybe there are no slots left!");
            return false;
        } else {
            return false;
        }
    }

    public string[] GetGameList() {
        if (m_connected) {
            return m_mmc.GetAdvertisements();
        } else {
            return null;
        }
    }

    public bool NewGameOnline(string map, string player) {
        if (m_connected) {
            string id = m_mmc.AdvertiseNewGame(map, player);
            if (id != null) {
                return JoinGame(id, player);
            }
            return false;
        } else {
            return false;
        }
    }

    public bool NewGameOffline(string map, string player) {
        NetworkGameObjectConnector.m_connectors.Clear();
        NetworkGameObjectConnector.m_humanRedableIDs.Clear();
        m_offlineGame = true;
        return true;
    }

    public GameInfo GetGameInfo(string id) {
        if (m_connected) {
            return m_mmc.GetGameInfo(id);
        } else {
            return new GameInfo();
        }
    }

    public GameInfo GetCurrentGameInfo() {
        if (m_connected && m_gc != null) {
            return m_mmc.GetGameInfo(m_gc.m_id);
        } else {
            return new GameInfo();
        }
    }

    void OnApplicationQuit() {
        if (m_gc != null)
            m_gc.Disconnect();
        if (m_connected)
            m_mmc.Disconnect();
        
    }

    // ATTRIBUTION: https://stackoverflow.com/questions/9184494/how-to-serialize-to-a-variable
    static byte[] SerializeArgument<T>(T argument) {
        byte[] bytes;
        using(var stream = new MemoryStream()) {
            var serializer = new DataContractSerializer(typeof(T));
            serializer.WriteObject(stream, argument);
            bytes = new byte[stream.Length];
            stream.Position = 0;
            stream.Read(bytes, 0, (int)stream.Length);
        }
        Debug.Log(String.Format("Serialized argument {0} to byte array with the size of {1}", argument.GetType().ToString(), bytes.Length.ToString()));
        return bytes;
    }

    // ATTRIBUTION: https://stackoverflow.com/questions/9184494/how-to-serialize-to-a-variable
    public static object[] DeserializeArgument(string message, Type[] t) {
        byte[] data = System.Text.Encoding.ASCII.GetBytes(message);

        object[] deserializedThing = default(object[]);

        using(var stream = new MemoryStream(data))
        using(var reader = XmlDictionaryReader.CreateTextReader(stream, new XmlDictionaryReaderQuotas()))
        {
            var serializer = new DataContractSerializer(typeof(object[]), t);

            // Deserialize the data and read it from the instance.
            deserializedThing = (object[])serializer.ReadObject(reader, true);
        }
        return deserializedThing;
    }
}
