﻿/*
AwesomeNET by xeetsh - Maximilian T. Schmitt

Licensed under the Unity Asset Store Terms of Service and EULA. See https://unity3d.com/legal/as_terms for a complete version of the license

AwesomeNET Doccumentation: https://docs.awesomenet.xeet.sh
Contact/Support/Feature requests: https://xeet.sh/contact.php
Twitter: @xeetsh
*/

using System.Collections;
using System.Collections.Generic;
using System.Net;
using System.Net.Sockets;
using System;
using UnityEngine;
using System.IO;

public enum MessageTypes {
    Error, Notification, Warning, ClientError
}


public abstract class TCPClient {
    public string m_host;
    public int m_port;

    TcpClient m_client;
    NetworkStream m_stream;
    public string[] m_lastArgv;

    public TCPClient(string host, int port) {
        m_host = host;
        m_port = port;
    }

    public bool Connect() {
        try {
            m_client = new TcpClient(m_host, m_port);
            m_stream = m_client.GetStream();

            if (AnalyzeMessage(Receive()) == (int) CCStatus.Success) {
                OnConnect();
                return true;
            }
            else {
                PrintMessage(string.Format("Error while connecting with with {0}:{1}", m_host, m_port), MessageTypes.Notification);
                return false;
            }

        } catch (SocketException) {
            PrintMessage(string.Format("Error while connecting with {0}:{1}", m_host, m_port), MessageTypes.Notification);
            return false;
        }
    }

    public bool Disconnect() {
        try {
            OnDisconnect();
            m_stream.Close();
            m_client.Close();
            return true;
        } catch (SocketException) {
            PrintMessage("Error while disconnecting", MessageTypes.Notification);
            return false;
        }
    }


    public string Receive() {
        try {
            byte[] data = new byte[8192];
            int bytes = m_stream.Read(data, 0, data.Length);
            string responseData = System.Text.Encoding.ASCII.GetString(data, 0, bytes);
            return responseData;
        } catch (Exception e) {
            // This throws a IOException if the gameclient disconnects while reading (which basically happens everytime a player quits its game). IOExceptions will be ignored
            if (e is SocketException) {
                PrintMessage("Error while listening to " + GetServerType(false) + "server", MessageTypes.Notification);
            }
            return CCStatus.Error.ToString();
        }
    }

    public void Send(int controlCode, string[] argv = null) {
        try {
            string message = controlCode.ToString();
            if (argv != null) {
                for(int i = 0; i < argv.Length; i++) {
                    if (argv[i] != "")
                        message += "_" + argv[i];
                }
            }

            byte[] data = System.Text.Encoding.ASCII.GetBytes(message);
            m_stream.Write(data, 0, data.Length);
        } catch (SocketException) {
            PrintMessage("Error while communicating with" + GetServerType(false) + "server", MessageTypes.Notification);
        }
    }

    public int AnalyzeMessage(string data) {
        int controlCode = 0;
        string[] argv = null;

        try {
            string[] dataA = data.Split('_');
            controlCode = int.Parse(dataA[0]);
            
            if (dataA.Length > 1) {
                argv = data.Substring(4).Split('_');
            } else {
                argv = null;
            }
            
        } catch (System.FormatException) {
            PrintMessage("Receive unclear instructions from " + GetServerType(false) + "server", MessageTypes.ClientError);
        }

        m_lastArgv = argv;

        return controlCode;
    }

    public void PrintMessage(string message, MessageTypes type) {
        string output = "[" + GetServerType(true) + "]";
            switch (type) {
                case MessageTypes.Notification:
                    output += " ";
                    break;

                case MessageTypes.Warning:
                    output += " [Warning] ";
                    break;

                case MessageTypes.Error:
                    output += " [Error] ";
                    break;

                case MessageTypes.ClientError:
                    output += " [Client Error] ";
                    break;
            }


            Debug.Log(output + message);
    }

    public abstract void OnConnect();
    public abstract void OnDisconnect();
    public abstract string GetServerType(bool uppercase);

}
